#include <autoconf.h>
#include <hermit.h>
#include <herrno.h>
#include <arch/system.h>
#include <io.h>
#include <htypes.h>
#include <string.h>
#include <delay.h>

extern int rcm2_uart_request_port(int port);

#define INVALID_CONSOLE (-1)
static int cur_console = INVALID_CONSOLE;

struct rcm2_uart {
	char *name;
	addr_t base;
	u32 mstp2_mask;
};

struct rcm2_uart uart[] = {
	{ .name = "ttySC0", .base = SCIFA0_BASE_ADDR,
	  .mstp2_mask = CPG_MSTP2_SCIFA0},
	{ .name = "ttySC1", .base = SCIFA1_BASE_ADDR,
	  .mstp2_mask = CPG_MSTP2_SCIFA1},
	{ .name = "ttySC2", .base = SCIFB0_BASE_ADDR,
	  .mstp2_mask = CPG_MSTP2_SCIFB0},
	{ .name = "ttySC3", .base = SCIFB1_BASE_ADDR,
	  .mstp2_mask = CPG_MSTP2_SCIFB1},
	{ .name = "ttySC4", .base = SCIFB2_BASE_ADDR,
	  .mstp2_mask = CPG_MSTP2_SCIFB2},
	{ .name = "ttySC5", .base = SCIFA2_BASE_ADDR,
	  .mstp2_mask = CPG_MSTP2_SCIFA2},
};

static int rcm2_uart_is_scifb(int port)
{
	if ((2 <= port) && (port <= 4))
		return 1;

	return 0;
}

static void rcm2_uart_supply_clock(int port)
{
	u32 sr, cr;
	u32 mask = uart[port].mstp2_mask;

	sr = read32(CPG_MSTPSR2);
	if (sr & mask) {
		cr = read32(CPG_SMSTPCR2);
		write32(CPG_SMSTPCR2, cr & ~mask);
	}
}

static void rcm2_setup_uart(int port)
{
	u16 status;

	rcm2_uart_supply_clock(port);

	/* wait transmission end */
	do {
		status = read16(uart[port].base + SCIF_FSR);
	} while (!(status & SCIF_FSR_TEND));

	/* uart disable */
	write16(uart[port].base + SCIF_SCR, 0x0000);

	/* reset transmit/receive FIFO */
	write16(uart[port].base + SCIF_FCR, 0x0006);

	/* set baudrate 115200 */
	write16(uart[port].base + SCIF_SMR, 0x0000);
	write8(uart[port].base + SCIF_BRR, 0x1b);

	/* Wait one bit interval */
	udelay((1000000+(115200-1)) / 115200);

	write16(uart[port].base + SCIF_FCR, 0x0000);

	/* uart enable */
	write16(uart[port].base + SCIF_SCR, 0x0030);
}

int change_console(char *console)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(uart); i++) {
		if (strncmp(console, uart[i].name, strlen(uart[i].name)) == 0
		    && rcm2_uart_request_port(i) == 0) {
			cur_console = i;
			if (uart[i].base)
				rcm2_setup_uart(i);
			return 0;
		}
	}

	cur_console = INVALID_CONSOLE;

	return -H_EINVAL;
}

static void rcm2_uart_handle_error(void)
{
	read16(uart[cur_console].base + SCIF_FSR);
	write16(uart[cur_console].base + SCIF_FSR, SCIF_FSR_CLR_ERR);
}

static int rcm2_uart_getchar_test(void)
{
	u16 fsr;

	fsr = read16(uart[cur_console].base + SCIF_FSR);
	if (fsr & SCIF_FSR_ERRS)
		rcm2_uart_handle_error();

	return fsr & (SCIF_FSR_DR | SCIF_FSR_RDF);
}

int arch_getchar(unsigned long timeout)
{
	u32 frdr;
	int ch;

	if (cur_console == INVALID_CONSOLE)
		return -H_EINVAL;

	if (rcm2_uart_is_scifb(cur_console))
		frdr = SCIFB_FRDR;
	else
		frdr = SCIFA_FRDR;

	while (!rcm2_uart_getchar_test());

	ch = read8(uart[cur_console].base + frdr);

	/* Dummy read */
	read16(uart[cur_console].base + SCIF_FSR);
	write16(uart[cur_console].base + SCIF_FSR, 0xfc);

	return ch;
}

int arch_putchar(int c, unsigned long timeout)
{
	u32 ftdr;

	if (cur_console == INVALID_CONSOLE)
		return -H_EINVAL;

	/* waiting for transmission end */
	while(1) {
		if (read16(uart[cur_console].base + SCIF_FSR) & SCIF_FSR_TEND)
			break;
	}

	if (rcm2_uart_is_scifb(cur_console))
		ftdr = SCIFB_FTDR;
	else
		ftdr = SCIFA_FTDR;

	write8(uart[cur_console].base + ftdr, c);
	write16(uart[cur_console].base + SCIF_FSR, 
		0x00df & ~SCIF_FSR_TEND);

	return 0;
}
